/*
 * taskset.c - taskset
 * Command-line utility for setting and retrieving a task's CPU affinity
 *
 * Robert Love <rml@tech9.net>		25 April 2002
 *
 * Linux kernels as of 2.5.8 provide the needed syscalls for
 * working with a task's cpu affinity.  Currently 2.4 does not
 * support these syscalls, but patches are available at:
 *
 * 	http://www.kernel.org/pub/linux/kernel/people/rml/cpu-affinity/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, v2, as
 * published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright (C) 2003 Robert Love
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <getopt.h>
//#include <sched.h>
#include <errno.h>

#include "include/taskset.h"
#include "include/version.h"

void show_usage(const char * cmd)
{
	printf("taskset version " VERSION "\n");
	printf("usage: %s [options] [mask] [pid | cmd [args...]]\n", cmd);
	printf("set or get the affinity of a process\n\n");
	printf("  -p, --pid              operate on existing given pid\n");
	printf("  -h, --help             display this help\n");
	printf("  -v, --version          output version information\n\n");
	printf("The default behavior is to run a new command:\n");
	printf("  %s 03 sshd -b 1024\n", cmd);
	printf("You can also retrieve the affinity of an existing task:\n");
	printf("  %s -p 700\n", cmd);
	printf("Or set it:\n");
	printf("  %s -p 03 700\n\n", cmd);
	printf("Report bugs and send patches to <rml@tech9.net>.\n");
}

int main(int argc, char * argv[])
{
	unsigned long new_mask;
	unsigned long cur_mask;
	unsigned int len = sizeof(new_mask);
	pid_t pid = 0;
	int i;

	struct option longopts[] = {
		{ "pid",	0, NULL, 'p' },
		{ "help",	0, NULL, 'h' },
		{ "version",	0, NULL, 'V' },
		{ NULL,		0, NULL, 0 }
	};

	while((i = getopt_long(argc, argv, "phV", longopts, NULL)) != -1) {
		switch(i) {
			case 'p':
				pid = atoi(argv[argc - 1]);
				break;
			case 'V':
				printf("taskset version " VERSION "\n");
				return 0;
			case ':':
			case '?':
			case 'h':
				show_usage(argv[0]);
				return 0;
		}
	}

	if ((!pid && argc - optind < 2)
			|| (pid && (argc - optind < 1 || argc - optind > 2))) {
		show_usage(argv[0]);
		return 1;
	}

	if (pid) {
		if (sched_getaffinity(pid, len, &cur_mask) < 0) {
			perror("sched_getaffinity");
			fprintf(stderr, "failed to get pid %d's affinity\n", pid);
			return 1;
		}
		printf("pid %d's current affinity mask: %08lx\n",pid, cur_mask);
		if (argc - optind == 1)
			return 0;
	}

	errno = 0;
	new_mask = strtoul(argv[optind], NULL, 16);
	if (errno) {
		perror("strtoul");
		fprintf(stderr, "failed to parse CPU mask %s\n", argv[optind]);
		return 1;
	}

	if (sched_setaffinity(pid, len, &new_mask)) {
		perror("sched_setaffinity");
		fprintf(stderr, "failed to set pid %d's affinity.\n", pid);
		return 1;
	}

	if (sched_getaffinity(pid, len, &cur_mask) < 0) {
		perror("sched_setaffinity");
		fprintf(stderr, "failed to get pid %d's affinity.\n", pid);
		return 1;
	}

	if (pid)
		printf("pid %d's new affinity mask: %08lx\n", pid, cur_mask);
	else {
		argv += optind + 1;
		execvp(argv[0], argv);
		perror("execvp");
		fprintf(stderr, "failed to execute %s\n", argv[0]);
		return 1;
	}

	return 0;
}
